@@ -8,10 +8,11 @@ import numpy as np
 import torch
 from sklearn import metrics
 from torch.nn import Module
-from torch.optim.lr_scheduler import MultiStepLR
+from torch.optim.lr_scheduler import ReduceLROnPlateau
 from torch.utils.data import DataLoader
 
-from pyg_extension.utils.log_model import LogModule, get_layers_with_weight, HookGradientModule, make_dot_
+from pyg_extension.utils.log_model import LogModule, HookGradientModule, make_dot_, \
+    get_target_layer, AverageMeterTotal
 
 
 def class_eval(prediction, target):
@@ -32,34 +33,6 @@ def class_eval(prediction, target):
     return accuracy, precision, recall, fscore, auc_score
 
 
-class AverageMeter(object):
-    """Computes and stores the average and current value"""
-
-    def __init__(self):
-        self.val = 0
-        self.avg = 0
-        self.sum = 0
-        self.count = 0
-
-    def reset(self):
-        self.val = 0
-        self.avg = 0
-        self.sum = 0
-        self.count = 0
-
-    def update(self, val, n=1):
-        self.val = val
-        self.sum = self.sum + val * n  # (if pytorch not support +=)
-        self.count += n
-        self.avg = self.sum / self.count
-
-
-def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
-    torch.save(state, filename)
-    if is_best:
-        shutil.copyfile(filename, 'model_best.pth.tar')
-
-
 def mae(prediction, target):
     """
     Computes the mean absolute error between prediction and target
@@ -72,14 +45,6 @@ def mae(prediction, target):
     return torch.mean(torch.abs(target - prediction))
 
 
-# def for_hook(module, input, output):
-#     print(module)
-#     for val in input:
-#         print("input val:", val)
-#     for out_val in output:
-#         print("output val:", out_val)
-
-
 class LearningFlow:
     """
     LearningFlow for training.
@@ -123,7 +88,7 @@ class LearningFlow:
         >>> device = torch.device('cuda:1' if torch.cuda.is_available() else 'cpu')
 
         >>> lf= LearningFlow(model, train_loader, validate_loader=val_loader, device= "cuda:1",
-        ... optimizer=None, clf= False, loss_method=None, learning_rate = 1e-3, milestones=None,
+        ... optimizer=None, clf= False, loss_method=None, learning_rate = 1e-3,
         ... weight_decay= 0.01, checkpoint=True, scheduler=scheduler,
         ... loss_threshold= 0.1, print_freq= None, print_what="all")
 
@@ -133,9 +98,10 @@ class LearningFlow:
     """
 
     def __init__(self, model: Module, train_loader: DataLoader, validate_loader: DataLoader, device="cpu",
-                 optimizer=None, clf: bool = False, loss_method=None, learning_rate: float = 1e-3, milestones=None,
+                 optimizer=None, clf: bool = False, loss_method=None, learning_rate: float = 1e-3,
                  weight_decay: float = 0.0, checkpoint=True, scheduler=None, debug: str = "", target_layers=(),
-                 loss_threshold: float = 0.1, print_freq: Union[int, str, None] = 10, print_what="all", process_label=None):
+                 loss_threshold: float = 0.1, print_freq: Union[int, str, None] = 10, print_what="all",
+                 process_label=None):
         """
 
         Parameters
@@ -152,8 +118,6 @@ class LearningFlow:
             see more in torch
         learning_rate:float
             see more in torch
-        milestones:list of float,None
-            see more in torch
         weight_decay:float
             see more in torch
         checkpoint:bool
@@ -180,7 +144,6 @@ class LearningFlow:
         self.model.to(device)
         self.clf = clf
         self.loss_method = loss_method
-        self.milestones = milestones
         self.optimizer = optimizer
         self.checkpoint = checkpoint
         self.weight_log = LogModule()
@@ -189,7 +152,7 @@ class LearningFlow:
 
         self.train_batch_number = len(self.train_loader)
 
-        if print_freq == "default" or print_freq is None:
+        if print_freq is None:
             self.print_freq = self.train_batch_number
         else:
             self.print_freq = self.train_batch_number if not isinstance(print_freq, int) else print_freq
@@ -214,10 +177,8 @@ class LearningFlow:
 
         else:
             self.loss_method = loss_method
-        if self.milestones is None:
-            self.milestones = [30, 50, 80]
         if scheduler is None:
-            self.scheduler = MultiStepLR(self.optimizer, milestones=self.milestones)
+            self.scheduler = ReduceLROnPlateau(self.optimizer, mode='min', factor=0.5, patience=7, min_lr=0.00001)
         else:
             self.optimizer = scheduler.optimizer
             self.scheduler = scheduler
@@ -230,40 +191,63 @@ class LearningFlow:
         self.forward_hook_list = []
         func = lambda x: x
         self._process_label = func if process_label is None else process_label
+        self.target_layers = target_layers
 
-        # hook
-        self.target_layers = ()
-        self.hook_layer(target_layers)
-
-    def hook_layer(self, target_layers):
-        if isinstance(target_layers, (list, tuple)) and len(target_layers) > 0:
-            self.target_layers = target_layers
-        elif isinstance(target_layers, str):
-            mod = []
-            if target_layers == "all":
-                # all layer
-                for name, module in self.model.named_modules():
-                    mod.append(module)
-            elif target_layers == "top":
-                # top layer
-                for name, module in self.model._modules.items():
-                    mod.append(module)
-            # just with weight layer
-            else:
-                mmdict = get_layers_with_weight(self.model)
-                for name, module in mmdict.items():
-                    mod.append(module)
-            self.target_layers = mod[:]
+        if self.clf:
+            self.meters = AverageMeterTotal("Time", "losses", "accuracies", "precisions", "recalls", "fscores",
+                                            "auc_scores")
         else:
-            mod = []
-            mmdict = get_layers_with_weight(self.model)
-            for name, module in mmdict.items():
-                mod.append(module)
-            self.target_layers = mod[:]
+            self.meters = AverageMeterTotal("Time", "losses", "mae_errors")
+
+    def my_debug(self):
+        # debug
+        if "loop" in self.debug:
+            self.weight_log.stats_loop()
+
+        if "single" in self.debug:
+            self.model.train()
+            for m, data in enumerate(self.train_loader):  # just record for once,
+                data = data.to(self.device)
+                self.model(data)
+                self.weight_log.get_weight(self.model, append=False)
+                break
+            self.weight_log.stats_single()
+
+        if "graphviz" in self.debug:
+
+            self.model.train()
+            for m, data in enumerate(self.train_loader):  # just record for once,
+                data = data.to(self.device)
+                y_pred = self.model(data)
+                vis_graph = make_dot_(y_pred, params=dict(list(self.model.named_parameters())))
+                vis_graph.render(format="pdf",view=False)
+                break
+
+        if "hook" in self.debug:
+            # which layer to hook.
+            self.target_layers = get_target_layer(self.model, self.target_layers)
+            self.gradient_log = HookGradientModule(self.target_layers)
+            self.model.train()
+            for m, data in enumerate(self.train_loader):  # just record for once,
+                data = data.to(self.device)
+                batch_y = self.process_label(data.y)
+                self.model.zero_grad()
+                y_pred = self.model(data)
+                lossi = self.loss_method(y_pred, batch_y)
+                lossi.backward(retain_graph=True)
+                # lossi.backward()
+                break
+            self.gradient_log.stats_single()
 
     def process_label(self, y):
         return self._process_label(y)
 
+    @staticmethod
+    def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
+        torch.save(state, filename)
+        if is_best:
+            shutil.copyfile(filename, 'model_best.pth.tar')
+
     def run(self, epoch=50, warm_start=False):
         """
         run loop.
@@ -284,7 +268,7 @@ class LearningFlow:
                 print("=> loading checkpoint '{}'".format(resume))
                 checkpoint = torch.load(resume)
                 start_epoch = checkpoint['epoch']
-                self.threshold = checkpoint['best_mae_error']
+                self.threshold = checkpoint['best_error']
                 self.model.load_state_dict(checkpoint['state_dict'])
                 self.optimizer.load_state_dict(checkpoint['optimizer'])
                 print("=> loaded checkpoint '{}' (epoch {})"
@@ -292,90 +276,54 @@ class LearningFlow:
             else:
                 print("=> no checkpoint found at '{}'".format(resume))
 
-        # self.model.to(self.device)
         if start_epoch > 0:
             epoch += start_epoch
             print("Try to run start from 'resumed epoch' {} to 'epoch' {}".format(start_epoch, epoch))
+
+        self.model.to(self.device)
+
         for epochi in range(start_epoch, epoch):
 
             self._train(epochi)
 
-            if "loop" in self.debug:
+            if "loop" in self.debug in self.debug:  # record each epochi
                 self.weight_log.get_weight(self.model)
 
-            if "single" in self.debug:
-                self.weight_log.stats()
-
             score = self._validate(epochi)
 
             if score != score:
                 print('Exit due to NaN')
                 sys.exit(1)
             try:
-                self.scheduler.step()
-            except TypeError:
                 self.scheduler.step(score)
+            except TypeError:
+                self.scheduler.step()
 
             is_best = score < self.best_error
             self.best_error = min(score, self.best_error)
 
             if self.checkpoint:
-                save_checkpoint({
+                self.save_checkpoint({
                     'epoch': epochi + 1,
                     'state_dict': self.model.state_dict(),
-                    'best_mae_error': self.threshold,
+                    'best_error': self.threshold,
                     'optimizer': self.optimizer.state_dict(),
                 }, is_best)
 
             if score <= self.threshold:
                 print("Up to requirements and early termination in epoch ({})".format(epochi))
                 break
-        # debug
 
-        if "loop" in self.debug:
-            self.weight_log.stats_loop()
-        if "graphviz" in self.debug:
-            self.model.train()
-            for m, data in enumerate(self.train_loader):
-                data = data.to(self.device)
-                y_pred = self.model(data)
-                vis_graph = make_dot_(y_pred, params=dict(list(self.model.named_parameters())))
-                vis_graph.render(format="pdf")
-                break
-        if "hook" in self.debug:
-            self.gradient_log = HookGradientModule(self.target_layers)
-            self.model.train()
-            for m, data in enumerate(self.train_loader):
-                data = data.to(self.device)
-                batch_y = self.process_label(data.y)
-                self.model.zero_grad()
-
-                y_pred = self.model(data)
-
-                lossi = self.loss_method(y_pred, batch_y)
-                lossi.backward(retain_graph=True)
-                # lossi.backward()
-                break
-            self.gradient_log.stats_single()
+        self.my_debug()
 
     def _train(self, epochi):
         self.model.train()
-        batch_time = AverageMeter()
-        losses = AverageMeter()
-        if self.clf is False:
-            mae_errors = AverageMeter()
-        else:
-            accuracies = AverageMeter()
-            precisions = AverageMeter()
-            recalls = AverageMeter()
-            fscores = AverageMeter()
-            auc_scores = AverageMeter()
+        self.meters.reset()
 
-        point = time.time()
         for m, data in enumerate(self.train_loader):
+            point = time.time()
             data = data.to(self.device)
             batch_y = self.process_label(data.y)
-            batch_time.update(time.time() - point)
 
             self.optimizer.zero_grad()
 
@@ -386,67 +334,42 @@ class LearningFlow:
             except TypeError:
                 target = batch_y.sign()
                 lossi = self.loss_method(y_pred, batch_y, target)
-            losses.update(float(lossi.cpu().item()), batch_y.size(0))
-
-            if self.clf is False:
-                mae_error = mae(y_pred.data.cpu(), batch_y.cpu())
-                mae_errors.update(mae_error, batch_y.size(0))
-            else:
-
-                accuracy, precision, recall, fscore, auc_score = \
-                    class_eval(y_pred.data.cpu(), batch_y.cpu())
-
-                accuracies.update(accuracy, batch_y.size(0))
-                precisions.update(precision, batch_y.size(0))
-                recalls.update(recall, batch_y.size(0))
-                fscores.update(fscore, batch_y.size(0))
-                auc_scores.update(auc_score, batch_y.size(0))
 
             lossi.backward()
 
             self.optimizer.step()
-            point = time.time()
 
-            if m % self.print_freq == 0 and self.print_what in ["all", "train"]:
-                if self.clf is False:
-                    print('Train: [{0}][{1}/{2}]\t'
-                          'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
-                          'MAE {mae_errors.val:.3f} ({mae_errors.avg:.3f})'
-                          'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'.format(epochi, m, len(self.train_loader),
-                                                                                      loss=losses,
-                                                                                      batch_time=batch_time,
-                                                                                      mae_errors=mae_errors))
-                else:
-                    print('Train: [{0}][{1}/{2}]\t'
-                          'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
-                          'Accu {accu.val:.3f} ({accu.avg:.3f})\t'
-                          'Precision {prec.val:.3f} ({prec.avg:.3f})\t'
-                          'Recall {recall.val:.3f} ({recall.avg:.3f})\t'
-                          'F1 {f1.val:.3f} ({f1.avg:.3f})\t'
-                          'AUC {auc.val:.3f} ({auc.avg:.3f})'
-                          'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'.format(
-                        epochi, m, len(self.train_loader), loss=losses,
-                        batch_time=batch_time,
-                        accu=accuracies,
-                        prec=precisions, recall=recalls, f1=fscores,
-                        auc=auc_scores))
+            # ###########
+
+            batch_size = batch_y.size(0)
+            self.meters['Time'].record(time.time() - point)
+            self.meters['losses'].record(float(lossi.cpu().item()), batch_size)
+            if self.clf is False:
+                mae_error = mae(y_pred.data.cpu(), batch_y.cpu())
+                self.meters['mae_errors'].record(mae_error, batch_size)
+            else:
+                accuracy, precision, recall, fscore, auc_score = \
+                    class_eval(y_pred.data.cpu(), batch_y.cpu())
+                self.meters['accuracies'].record(accuracy, batch_size)
+                self.meters['precisions'].record(precision, batch_size)
+                self.meters['recalls'].record(recall, batch_size)
+                self.meters['fscores'].record(fscore, batch_size)
+                self.meters['auc_scores'].record(auc_score, batch_size)
+
+            if self.print_freq != self.train_batch_number and m % self.print_freq == 0 \
+                    and self.print_what in ["all", "train"]:
+                print('Train: [{0}][{1}/{2}] {3}'.format(epochi, m, len(self.train_loader), self.meters.text()))
+
+        if self.print_freq == self.train_batch_number:
+            print('Train: [{}] {}'.format(epochi, self.meters.text()))
 
     def _validate(self, epochi):
         self.model.eval()
-        mae_errors = AverageMeter()
-        if self.clf is False:
-
-            losses = AverageMeter()
-        else:
-            accuracies = AverageMeter()
-            precisions = AverageMeter()
-            recalls = AverageMeter()
-            fscores = AverageMeter()
-            auc_scores = AverageMeter()
-
-        self.model.eval()
+        self.meters.reset()
 
         for m, data in enumerate(self.test_loader):
+            point = time.time()
+
             data = data.to(self.device)
             batch_y = self.process_label(data.y)
             y_pred = self.model(data)
@@ -457,54 +380,43 @@ class LearningFlow:
                 target = batch_y.sign()
                 lossi = self.loss_method(y_pred, batch_y, target)
 
-            losses.update(lossi.cpu().item(), batch_y.size(0))
-            if self.clf is False:
+            # ###########
 
-                mae_error = mae((y_pred.data.cpu()), batch_y.cpu())
-                mae_errors.update(mae_error, batch_y.size(0))
+            batch_size = batch_y.size(0)
+            self.meters['Time'].record(time.time() - point)
+            self.meters['losses'].record(float(lossi.cpu().item()), batch_size)
+            if self.clf is False:
+                mae_error = mae(y_pred.data.cpu(), batch_y.cpu())
+                self.meters['mae_errors'].record(mae_error, batch_size)
             else:
                 accuracy, precision, recall, fscore, auc_score = \
                     class_eval(y_pred.data.cpu(), batch_y.cpu())
+                self.meters['accuracies'].record(accuracy, batch_size)
+                self.meters['precisions'].record(precision, batch_size)
+                self.meters['recalls'].record(recall, batch_size)
+                self.meters['fscores'].record(fscore, batch_size)
+                self.meters['auc_scores'].record(auc_score, batch_size)
 
-                accuracies.update(accuracy, batch_y.size(0))
-                precisions.update(precision, batch_y.size(0))
-                recalls.update(recall, batch_y.size(0))
-                fscores.update(fscore, batch_y.size(0))
-                auc_scores.update(auc_score, batch_y.size(0))
+            if self.print_freq != self.train_batch_number and m % self.print_freq == 0 \
+                    and self.print_what in ["all", "train"]:
+                print('Test: [{0}][{1}/{2}] {3}'.format(epochi, m, len(self.train_loader), self.meters.text()))
 
-            if m % self.print_freq == 0 and self.print_what in ["all", "test"]:
-                if self.clf is False:
-                    print('Test: [{0}][{1}/{2}]\t'
-                          'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
-                          'MAE {mae_errors.val:.3f} ({mae_errors.avg:.3f})'.format(
-                        epochi, m, len(self.test_loader),
-                        loss=losses, mae_errors=mae_errors))
-                else:
-                    print('Test: [{0}][{1}/{2}]\t'
-                          'Loss {loss.val:.4f} ({loss.avg:.4f})\t'
-                          'Accu {accu.val:.3f} ({accu.avg:.3f})\t'
-                          'Precision {prec.val:.3f} ({prec.avg:.3f})\t'
-                          'Recall {recall.val:.3f} ({recall.avg:.3f})\t'
-                          'F1 {f1.val:.3f} ({f1.avg:.3f})\t'
-                          'AUC {auc.val:.3f} ({auc.avg:.3f})'.format(
-                        epochi, m, len(self.test_loader),
-                        loss=losses, accu=accuracies,
-                        prec=precisions, recall=recalls, f1=fscores,
-                        auc=auc_scores))
+        if self.print_freq == self.train_batch_number:
+            print('Test: [{}] {}'.format(epochi, self.meters.text()))
 
         if self.clf is False:
-            return mae_errors.avg
+            return self.meters['mae_errors'].avg
         else:
-            return auc_scores.avg
+            return self.meters['auc_scores'].avg
 
-    def mae_score(self, predict_loader):
+    def score(self, predict_loader):
         """Return MAE score."""
-        y_pre, y_true = self.predict(predict_loader, return_y_true=True, add_hook=False)
+        y_pre, y_true = self.predict(predict_loader)
         return float(mae(y_pre, y_true))
 
     def predict(self, predict_loader: DataLoader, return_y_true=False, add_hook=False, hook_layer=None):
         """
-        Just predict by model,and add one forward hook to get put.
+        Just predict by model,and add one forward hook to get processing output.
 
         Parameters
         ----------
@@ -514,7 +426,7 @@ class LearningFlow:
             if return_y_true, return (y_preds, y_true)
         add_hook:bool
         hook_layer:
-            layer of model.
+            layer of model. could get layers by ``get_target_layer`` or ``get_layers_name_with_weight``.
 
         Returns
         -------
